import asyncio
import json
import time
import traceback

from pydantic import BaseModel

from gravybox.betterstack import collect_logger
from gravybox.exceptions import GravyboxException
from gravybox.protocol import LinkRequest

logger = collect_logger()


def upstream_api_call(upstream_provider):
    """
    wrapper for all upstream api calls
    handles errors, task cancellations, metrics, and logging
    """

    def decorator(function):
        async def wrapper(*args, link_request: LinkRequest = None, **kwargs):
            if link_request is None:
                raise ValueError("please pass the original link request when making a call to an upstream api")
            call_args = [arg for arg in args]
            call_kwargs = [f"{key}={value}" for key, value in kwargs.items()]
            log_extras = {
                "upstream_provider": upstream_provider,
                "upstream_call_type": function.__name__,
                "upstream_call_arguments": json.dumps(call_args + call_kwargs),
                "trace_id": link_request.trace_id
            }
            logger.info("( ) calling upstream api", extra=log_extras)
            start_time = time.time()
            try:
                result: BaseModel = await function(*args, log_extras=log_extras, **kwargs)
                log_extras["elapsed_time"] = time.time() - start_time
                log_extras["result"] = result.model_dump_json()
                logger.info("(*) calling upstream api succeeded", extra=log_extras)
                return result
            except asyncio.CancelledError:
                logger.info("(*) calling upstream api cancelled, exiting gracefully", extra=log_extras)
                raise
            except Exception as error:
                if isinstance(error, GravyboxException):
                    log_extras |= error.log_extras
                log_extras["error_str"] = str(error)
                log_extras["traceback"] = traceback.format_exc()
                log_extras["elapsed_time"] = time.time() - start_time
                logger.warning("(!) calling upstream api failed", extra=log_extras)
                return None

        return wrapper

    return decorator



import pytest
from fastapi import FastAPI, APIRouter
from pydantic import BaseModel
from starlette.testclient import TestClient

from gravybox.endpoints import LinkEndpoint
from gravybox.exceptions import DataUnavailable
from gravybox.protocol import LinkRequest, LinkResponse

app = FastAPI()
app.add_middleware(LinkEndpoint)

router = APIRouter(prefix="/test")


class TestModel(BaseModel):
    value_one_plus_one: int
    opposite_of_value_two: bool


class TestRequest(LinkRequest):
    value_one: int
    value_two: bool


class TestResponse(LinkResponse):
    content: TestModel | None


@router.post("/success")
async def success_endpoint(link_request: TestRequest) -> TestResponse:
    result = TestModel(
        value_one_plus_one=link_request.value_one + 1,
        opposite_of_value_two=not link_request.value_two
    )
    return TestResponse(success=True, content=result)


@router.post("/failure")
async def failing_endpoint(link_request: TestRequest) -> TestResponse:
    raise RuntimeError("failing endpoint failed as expected")


@router.post("/data_unavailable")
async def data_unavailable_endpoint(link_request: TestRequest) -> TestResponse:
    raise DataUnavailable()


app.include_router(router)


@pytest.mark.asyncio
async def test_link_endpoint_no_payload():
    with TestClient(app) as client:
        response = client.post("/test/success")
        assert response.status_code == 400
        test_response = TestResponse.model_validate(response.json())
        assert test_response.success is False
        assert test_response.error == "request does not contain valid json, or is missing a trace_id"
        assert test_response.content is None


@pytest.mark.asyncio
async def test_link_endpoint_no_trace_id():
    with TestClient(app) as client:
        test_request = {
            "value_one": 3,
            "value_two": False
        }
        response = client.post("/test/success", json=test_request)
        assert response.status_code == 400
        test_response = TestResponse.model_validate(response.json())
        assert test_response.success is False
        assert test_response.error == "request does not contain valid json, or is missing a trace_id"
        assert test_response.content is None


@pytest.mark.asyncio
async def test_link_endpoint_malformed_payload():
    with TestClient(app) as client:
        test_request = {
            "trace_id": "malformed_payload",
            "value_one": 3
        }
        response = client.post("/test/success", json=test_request)
        assert response.status_code == 422


@pytest.mark.asyncio
async def test_link_endpoint_success():
    with TestClient(app) as client:
        test_request = TestRequest(trace_id="success", value_one=3, value_two=False)
        response = client.post("/test/success", json=test_request.model_dump())
        assert response.status_code == 200
        test_response = TestResponse.model_validate(response.json())
        assert test_response.success is True
        assert test_response.error == ""
        assert test_response.content.value_one_plus_one == 4
        assert test_response.content.opposite_of_value_two is True


@pytest.mark.asyncio
async def test_link_endpoint_failure():
    with TestClient(app) as client:
        test_request = TestRequest(trace_id="failure", value_one=3, value_two=False)
        response = client.post("/test/failure", json=test_request.model_dump())
        assert response.status_code == 500
        test_response = TestResponse.model_validate(response.json())
        assert test_response.success is False
        assert test_response.error == "server encountered unhandled exception"
        assert test_response.content is None


@pytest.mark.asyncio
async def test_link_endpoint_data_unavailable():
    with TestClient(app) as client:
        test_request = TestRequest(trace_id="data unavailable", value_one=3, value_two=False)
        response = client.post("/test/data_unavailable", json=test_request.model_dump())
        assert response.status_code == 500
        test_response = TestResponse.model_validate(response.json())
        assert test_response.success is False
        assert test_response.error == "data unavailable"
        assert test_response.content is None



import json
import time
import traceback

from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import JSONResponse, Response

from gravybox.betterstack import collect_logger
from gravybox.exceptions import DataUnavailable, GravyboxException
from gravybox.protocol import LinkResponse

logger = collect_logger()


class LinkEndpoint(BaseHTTPMiddleware):
    def __init__(self, app):
        super().__init__(app)

    async def dispatch(self, request, call_next):
        log_extras = {}
        try:
            payload = await request.json()
            log_extras["request_json"] = json.dumps(payload)
            log_extras["trace_id"] = payload["trace_id"]
        except Exception as error:
            log_extras["error_str"] = str(error)
            logger.error("(!) failed to parse request", extra=log_extras)
            return JSONResponse(
                status_code=400,
                content=LinkResponse(
                    success=False,
                    error="request does not contain valid json, or is missing a trace_id"
                ).model_dump()
            )
        logger.info("( ) link request", extra=log_extras)
        start_time = time.time()
        try:
            response: Response = await call_next(request)
            log_extras["elapsed_time"] = time.time() - start_time
            log_extras["status_code"] = response.status_code
            logger.info("(*) link response", extra=log_extras)
            return response
        except DataUnavailable:
            log_extras["elapsed_time"] = time.time() - start_time
            logger.warning("(!) could not find requested data", extra=log_extras)
            return JSONResponse(
                status_code=500,
                content=LinkResponse(
                    success=False,
                    error="data unavailable"
                ).model_dump()
            )
        except Exception as error:
            if isinstance(error, GravyboxException):
                log_extras |= error.log_extras
            log_extras["error_str"] = str(error)
            log_extras["traceback"] = traceback.format_exc()
            log_extras["elapsed_time"] = time.time() - start_time
            logger.error("(!) failed with unhandled exception", extra=log_extras)
            return JSONResponse(
                status_code=500,
                content=LinkResponse(
                    success=False,
                    error="server encountered unhandled exception"
                ).model_dump()
            )



import pytest

from gravybox.betterstack import collect_logger
from gravybox.exceptions import DataUnavailable
from gravybox.protocol import LinkRequest
from gravybox.upstream_centrifuge import UpstreamCentrifuge
from test.testkit import sleeping_coroutine, TestTaskResult, none_result_coroutine, failing_coroutine

logger = collect_logger()


@pytest.mark.asyncio
async def test_upstream_centrifuge_single_task():
    link_request = LinkRequest(trace_id="centrifuge")
    tasks = [
        sleeping_coroutine(1, "test", 23, field_three=True, link_request=link_request)
    ]
    centrifuge = UpstreamCentrifuge(tasks, TestTaskResult)
    result = await centrifuge.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_centrifuge_two_task():
    link_request = LinkRequest(trace_id="double_centrifuge")
    tasks = [
        sleeping_coroutine(1, "test", 23, field_three=True, link_request=link_request),
        sleeping_coroutine(999, "sleepy", 333, field_three=False, link_request=link_request)
    ]
    centrifuge = UpstreamCentrifuge(tasks, TestTaskResult)
    result = await centrifuge.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_centrifuge_failing_task():
    link_request = LinkRequest(trace_id="failing_centrifuge")
    tasks = [
        sleeping_coroutine(5, "test", 23, field_three=True, link_request=link_request),
        failing_coroutine(1, "failure", 333, field_three=False, link_request=link_request)
    ]
    centrifuge = UpstreamCentrifuge(tasks, TestTaskResult)
    result = await centrifuge.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_centrifuge_none_result_task():
    link_request = LinkRequest(trace_id="failing_centrifuge")
    tasks = [
        sleeping_coroutine(5, "test", 23, field_three=True, link_request=link_request),
        none_result_coroutine(2, None, None, field_three=False, link_request=link_request)
    ]
    centrifuge = UpstreamCentrifuge(tasks, TestTaskResult)
    result = await centrifuge.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_centrifuge_total_failure():
    link_request = LinkRequest(trace_id="total_failure")
    tasks = [
        failing_coroutine(2, "failure", 333, field_three=False, link_request=link_request),
        failing_coroutine(1, "failure", 333, field_three=False, link_request=link_request)
    ]
    centrifuge = UpstreamCentrifuge(tasks, TestTaskResult)
    with pytest.raises(DataUnavailable):
        await centrifuge.activate()


@pytest.mark.asyncio
async def test_upstream_centrifuge_merge_two_results():
    link_request = LinkRequest(trace_id="centrifuge_merge")
    tasks = [
        sleeping_coroutine(1, None, 23, field_three=True, link_request=link_request),
        sleeping_coroutine(3, "test", None, field_three=True, link_request=link_request),
    ]
    centrifuge = UpstreamCentrifuge(tasks, TestTaskResult)
    result = await centrifuge.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_centrifuge_merge_overwrite():
    link_request = LinkRequest(trace_id="merge_overwrite")
    tasks = [
        sleeping_coroutine(1, None, None, field_three=True, link_request=link_request),
        sleeping_coroutine(3, "test", None, field_three=True, link_request=link_request),
        sleeping_coroutine(5, "late_precedence", 15, field_three=False, link_request=link_request),
    ]
    centrifuge = UpstreamCentrifuge(tasks, TestTaskResult)
    result = await centrifuge.activate()
    assert result == TestTaskResult(field_one="late_precedence", field_two=15, field_three=False)


@pytest.mark.asyncio
async def test_upstream_centrifuge_chaos():
    link_request = LinkRequest(trace_id="centrifuge_chaos")
    tasks = [
        sleeping_coroutine(99, None, None, field_three=True, link_request=link_request),
        failing_coroutine(33, "failure", 333, field_three=False, link_request=link_request),
        sleeping_coroutine(1, None, None, field_three=True, link_request=link_request),
        none_result_coroutine(3, None, None, field_three=False, link_request=link_request),
        failing_coroutine(1, "failure", 333, field_three=False, link_request=link_request),
        sleeping_coroutine(3, "test", None, field_three=True, link_request=link_request),
        none_result_coroutine(2, None, None, field_three=False, link_request=link_request),
        failing_coroutine(2, "failure", 333, field_three=False, link_request=link_request),
        sleeping_coroutine(5, "late_precedence", 15, field_three=False, link_request=link_request),
        sleeping_coroutine(19, "something_else", 25, field_three=False, link_request=link_request)
    ]
    centrifuge = UpstreamCentrifuge(tasks, TestTaskResult)
    result = await centrifuge.activate()
    assert result == TestTaskResult(field_one="late_precedence", field_two=15, field_three=False)



from pydantic import BaseModel


class GravyboxRequest(BaseModel):
    trace_id: str


class GravyboxResponse(BaseModel):
    success: bool
    error: str = ""
    content: dict | None = None


class LinkRequest(GravyboxRequest):
    pass


class LinkResponse(GravyboxResponse):
    pass



import os

from gravybox.betterstack import collect_logger
from gravybox.exceptions import BadStatusCode
from gravybox.requests import AsyncRequestManager

logger = collect_logger()

RAPIDAPI_KEY = os.getenv("RAPIDAPI_KEY")


async def query_rapidapi(url, host, query, extra_headers=None):
    headers = {
        "X-RapidAPI-Key": RAPIDAPI_KEY,
        "X-RapidAPI-Host": host
    }
    if extra_headers is not None:
        headers |= extra_headers
    response = await AsyncRequestManager.client().get(url, headers=headers, params=query)
    if response.status_code == 200:
        return response.json()
    else:
        raise BadStatusCode(response)
